#include <SoftwareSerial.h>
#include "TimerOne.h"
#include <Wire.h>

// Master Code to sending a request to slave to start collectiong a scan results during drive 
#define SLAVE_ADDRESS 0x08
#define START_SIGNAL 0x01
SoftwareSerial bluetoothSerial(4, 5);  // RX, TX pins for Bluetooth module comunication

const byte SENSOR_A_PIN = 2;  // Sensor L393 A pin connected to Arduino digital pin 2
const byte SENSOR_B_PIN = 3;  // Sensor L393 B pin connected to Arduino digital pin 3

const int enA = 6;  // L298N pin out 
const int in1 = 7;
const int in2 = 8;
const int enB = 9;
const int in3 = 10;
const int in4 = 11;

// motor parameters
int mspeed = 125;// Motor speed 
int motor_B_speed_adjustment = 5;  // Number of slots to reduce the speed of motor B in testing process as one of the motor is slower then other
int motor_A_speed_adjustment = 5;// Number of slots to reduce the speed of motor A
volatile int counter_A = 0;  // Pulse counter for Motor A
volatile int counter_B = 0;  // Pulse counter for Motor B
const float CALIBRATION_FACTOR_LINEAR = 3.34; //(34:1 gearbox)factor to adjust values of oparator 
const float DISK_SLOTS = 20; //encoder disk slots number 
const float WHEEL_DIAMETER = 60.00;//diameter of motorised wheel
const float WHEEL_CIRCUMFERENCE = PI * WHEEL_DIAMETER;// circumference of the wheel approx 18.9 cm

int steps = 0;// Global variable declaration of steps 

// Scaner for obstcale detection 
const int trigPin = A1;// trig pin connected to v5 shield 
const int echoPin = A0;// echo pin connected to v5 shield 
float duration, distance;// HC 04 - variables to measure distance  

// Area for scanning 
int receivedDesignatedLength = 0;  // Variable to store received designated length
int receivedDesignatedWidth = 0;   // Variable to store received designated width
long designated_length = 0;  // Designated length as an integer input by oparator
long designated_width = 0;   // Designated width as an integer input by oparator
int vehicle_width = 20; // The vehicle's fixed Width (20 cm)
long total_area = 0 ;   // total area for scan  
int all_lanes = 0;  // Global variable declaration
bool waitingForDesignatedLength = true;  // Flag to indicate waiting for designated length
bool waitingForDesignatedWidth = true;  // Flag to indicate waiting for designated length

int CMtoSteps(float cm) { // cm to steps function
  int result=steps;  // Final calculation result 
  float circumference = WHEEL_DIAMETER * PI / 10;
  float cm_step = circumference / DISK_SLOTS;  
  float f_result = cm / cm_step; 
  result = round(f_result);
  return result;  
}
void ISR_countA() {
  counter_A++;  // Increment Motor A counter value
}
void ISR_countB() {
  counter_B++;  // Increment Motor B counter value
}

void moveForward(int steps, int mspeed){ // Function to Move in Reverse{
   counter_A = 0;  //  reset counter A to zero
   counter_B = 0;  //  reset counter B to zero
   
   digitalWrite(in1, HIGH);// Set Motor A forward
   digitalWrite(in2, LOW);
   digitalWrite(in3, HIGH);// Set Motor B forward
   digitalWrite(in4, LOW);
   
   // Go forward until step value is reached
   while (steps > counter_A && steps > counter_B) {
   
    if (steps > counter_A) {
    analogWrite(enA, mspeed);
    } else {
    analogWrite(enA, 0);
    }
    if (steps > counter_B) {
    analogWrite(enB, mspeed);
    } else {
    analogWrite(enB, 0);
    }
   }
    
  // Stop when done
  analogWrite(enA, 0);
  analogWrite(enB, 0);
  counter_A = 0;  //  reset counter A to zero
  counter_B = 0;  //  reset counter B to zero 
 
}
void moveBackward(int steps, int mspeed) // Function to Move in Reverse
{
   counter_A = 0;  //  reset counter A to zero
   counter_B = 0;  //  reset counter B to zero
   
  digitalWrite(in1, LOW);// Set Motor A reverse
  digitalWrite(in2, HIGH);
  digitalWrite(in3, LOW);// Set Motor B reverse
  digitalWrite(in4, HIGH);
   
   // Go in reverse until step value is reached
   while (steps > counter_A && steps > counter_B) {
   
    if (steps > counter_A) {
    analogWrite(enA, mspeed);
    } else {
    analogWrite(enA, 0);
    }
    if (steps > counter_B) {
    analogWrite(enB, mspeed);
    } else {
    analogWrite(enB, 0);
    }
    }
    
  // Stop when done
  analogWrite(enA, 0);
  analogWrite(enB, 0);
  counter_A = 0;  //  reset counter A to zero
  counter_B = 0;  //  reset counter B to zero 
 
}
void Stop() {// Stop Function to halt the motors
    analogWrite(enA, 0);  // Stop motor A
    analogWrite(enB, 0);  // Stop motor B
}
void turnRight(int steps, int mspeed) {
    counter_A = 0;
    counter_B = 0;

    digitalWrite(in1, LOW);
    digitalWrite(in2, HIGH);
    digitalWrite(in3, HIGH);
    digitalWrite(in4, LOW);
    
    float kp = 0.2;  // Proportional gain. This value might need tuning.

    while (steps > counter_A || steps > counter_B) {
        int error = counter_A - counter_B;

        int speedAdjustment = kp * error;
        int speedA = mspeed - speedAdjustment;
        int speedB = mspeed + speedAdjustment;

        // Limit speed to maximum value of mspeed.
        speedA = constrain(speedA, 0, mspeed);
        speedB = constrain(speedB, 0, mspeed);

        if (steps > counter_A) {
            analogWrite(enA, speedA);
        } else {
            analogWrite(enA, 0);
        }
        if (steps > counter_B) {
            analogWrite(enB, speedB);
        } else {
            analogWrite(enB, 0);
        }
    }

    analogWrite(enA, 0);
    analogWrite(enB, 0);
    counter_A = 0;
    counter_B = 0;
}
void turnLeft(int steps, int mspeed) // Function to Move left
{
   counter_A = 0;  //  reset counter A to zero
   counter_B = 0;  //  reset counter B to zero
   
  digitalWrite(in1, HIGH);// Set Motor A forward
  digitalWrite(in2, LOW);
  digitalWrite(in3, LOW); // Set Motor B reverse
  digitalWrite(in4, HIGH);
   
   // Go until step value is reached
   while (steps > counter_A && steps > counter_B) {
   
    if (steps > counter_A) {
    analogWrite(enA, mspeed);
    } else {
    analogWrite(enA, 0);
    }
    if (steps > counter_B) {
    analogWrite(enB, mspeed);
    } else {
    analogWrite(enB, 0);
   }
  }
    // Stop when done
  analogWrite(enA, 0);
  analogWrite(enB, 0);
  counter_A = 0;  //  reset counter A to zero
  counter_B = 0;  //  reset counter B to zero 
 
}    
void measureDistanceAndSendSerial() {// HC 04 function 
  digitalWrite(trigPin, LOW);
  delayMicroseconds(2);
  digitalWrite(trigPin, HIGH);
  delayMicroseconds(10);
  digitalWrite(trigPin, LOW);
  duration = pulseIn(echoPin, HIGH);
  distance = (duration * 0.0343) / 2;

  Serial.print("Distance:");// Send the distance measurement via serial
  Serial.println(distance);
}
void avoidObstacle() {// Obstacle detected, take avoidance action
  measureDistanceAndSendSerial();
  
  if (distance < 20) {
    Serial.println("Obstacle detected! Avoiding...");// Obstacle detected, turn on LED
    digitalWrite(LED_BUILTIN, HIGH);
    // Turn right 45* degree ,foward lenght of vehicle ,Turn left 45* degree,foward lenght of vehicle,Turn left 45* degree,Turn right 45* degree
    turnRight(40,mspeed); moveForward(67,mspeed); turnLeft(40,mspeed); moveForward(67,mspeed);turnLeft(40,mspeed);moveForward(67,mspeed);turnRight(40,mspeed);
    
  }
}
void processBluetoothData() {// input string at the comma to get designated_length and designated_width
  if (bluetoothSerial.available()) {
    String input = bluetoothSerial.readStringUntil('\n');
    
    // Split the input string at the comma to get designated_length and designated_width
    int commaIndex = input.indexOf(',');
    if (commaIndex != -1) {
      receivedDesignatedLength = input.substring(0, commaIndex).toFloat();
      receivedDesignatedWidth = input.substring(commaIndex + 1).toFloat();
      waitingForDesignatedLength = false;
      waitingForDesignatedWidth = false;
      measureDistanceAndSendSerial();

    }
  }
}
void setup() { 
  
  pinMode(enA, OUTPUT);  // Motor connection
  pinMode(in1, OUTPUT);
  pinMode(in2, OUTPUT);
  pinMode(enB, OUTPUT);
  pinMode(in3, OUTPUT);
  pinMode(in4, OUTPUT); 
  pinMode(trigPin, OUTPUT);// HC04 Trig pin
  pinMode(echoPin, INPUT);//  HC04 Echo pin
  pinMode(LED_BUILTIN, OUTPUT);  // Use the built-in LED on pin 13
  Wire.begin();  // Initialize I2C communications as Master
  Serial.begin(9600);//serial communication for debugging
  bluetoothSerial.begin(9600);  // Initialize Bluetooth serial communication at 9600 baud
  
  attachInterrupt(digitalPinToInterrupt(SENSOR_A_PIN), ISR_countA, RISING);
  attachInterrupt(digitalPinToInterrupt(SENSOR_B_PIN), ISR_countB, RISING);
}
void moveToNextLaneOdd() {//swich lane to right -- move 90* right move length of lane and turn right 90* to the next lane
      Stop();delay(500);
      turnRight(60,mspeed);  Serial.println("rrr");
      moveForward(60,mspeed);Serial.println("ffffff");
      turnRight(90,mspeed);Serial.println("rrrrr");
  
}
void moveToNextLaneEven() {//swich lane to left -- move 90* left move length of lane and turn left 90* to the next lane
       turnLeft(80,mspeed);Stop();delay(500);
       moveForward(67,mspeed);
       turnLeft(80,mspeed);
   

}
void handleBluetoothCommands() {
    String input = bluetoothSerial.readStringUntil('\n');

    // Manual mode as option to drive rover manually  
    if (input.charAt(0) == 'F') {Serial.println("Forward");
        moveForward(334, mspeed);
    } else if (input.charAt(0) == 'B') {Serial.println("Backward");
        moveBackward(334, mspeed);
    } else if (input.charAt(0) == 'R') {Serial.println("Right");
        turnRight(80, mspeed);
    } else if (input.charAt(0) == 'L') {Serial.println("Left");
        turnLeft(80, mspeed);
    } else if (input.charAt(0) == 'S') {
        Stop();
    } else {
       
        int commaIndex = input.indexOf(',');// Handle the designated length and width input in case when the manual is not used
        if (commaIndex != -1) {
            designated_length = input.substring(0, commaIndex).toInt();
            designated_width = input.substring(commaIndex + 1).toInt();
            waitingForDesignatedLength = false;
            waitingForDesignatedWidth = false;
        }
    }
}
void executeAutomaticMode() {//Starting automatic mode
    Wire.beginTransmission(SLAVE_ADDRESS);// Start communication with slave
    Wire.write(START_SIGNAL);
    total_area = abs((designated_width*CALIBRATION_FACTOR_LINEAR) * (designated_length*CALIBRATION_FACTOR_LINEAR));
    int all_lanes = designated_width / vehicle_width;Serial.println(all_lanes);
    int op_input = designated_length*CALIBRATION_FACTOR_LINEAR;// oparator input value
    Wire.write((byte)all_lanes);
    Wire.endTransmission();
    
    for (int lane = 0; lane < all_lanes; lane++) {
        measureDistanceAndSendSerial();
        avoidObstacle();
        moveForward((op_input), mspeed);
        delay(500);
        
        if (lane % 2 == 0) {
            moveToNextLaneEven();
            delay(500);
            Serial.println("even turn");
        } else {
            moveToNextLaneOdd();
            delay(500);
            Serial.println("ood turn");
        }  
    }
    
    
 
  
    
    delay(50);
   // counter_A = 0;
   // counter_B = 0;
    //waitingForDesignatedLength = true;
   // waitingForDesignatedWidth = true;
}
float designatedLength = 0;  
float designatedWidth = 0;
void sendRequestToSlave() {
    Serial.println("Final request to slave after all lanes processed.");
    Wire.requestFrom(SLAVE_ADDRESS, 300);
    while (Wire.available()) {
        char c = Wire.read();
        bluetoothSerial.print(counter_A);
        bluetoothSerial.print(",");
        bluetoothSerial.print(counter_B);
        bluetoothSerial.print(",");
        bluetoothSerial.print(all_lanes); // Send the area back to the phone via Bluetooth
    bluetoothSerial.println();
    delay(50);  // Delay after completing the request
    }}
void loop() {
    
    processBluetoothData();// Process any incoming Bluetooth data

    if (bluetoothSerial.available()) {  // Check if there's data available to read from app
        handleBluetoothCommands();// Manual control if needed
    }

    // If  received the designated length and width, execute the automatic mode logic
    if (!waitingForDesignatedLength && !waitingForDesignatedWidth) {
        executeAutomaticMode();
          sendRequestToSlave();  // Here you call the request logic
        // Reset counters and state
        counter_A = 0;
        counter_B = 0;
        waitingForDesignatedLength = true;
        waitingForDesignatedWidth = true;
    }
  
}